/**
 * Copyright (c) 2016-2018 Thales Global Services S.A.S.
 * This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 2.0
 * which is available at https://www.eclipse.org/legal/epl-2.0/
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *   Abel Hegedus, Tamas Borbas, Balazs Grill, Peter Lunk, Daniel Segesdi (IncQuery Labs Ltd.) - initial API and implementation
 */
package org.eclipse.emf.diffmerge.patch.runtime;

import com.google.common.base.Objects;
import com.google.common.base.Optional;
import com.google.common.base.Preconditions;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import java.util.List;
import java.util.Map;
import org.eclipse.emf.diffmerge.patch.AttributeEntry;
import org.eclipse.emf.diffmerge.patch.ElementEntry;
import org.eclipse.emf.diffmerge.patch.ReferenceEntry;
import org.eclipse.emf.diffmerge.patch.api.ChangeDirection;
import org.eclipse.emf.diffmerge.patch.api.EntryType;
import org.eclipse.emf.diffmerge.patch.api.ModelPatch;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchDiagnosticElement;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchEntry;
import org.eclipse.emf.diffmerge.patch.api.PatchApplication;
import org.eclipse.emf.diffmerge.patch.api.PatchApplicationDiagnostic;
import org.eclipse.emf.diffmerge.patch.runtime.FeatureHolder;
import org.eclipse.emf.diffmerge.patch.runtime.identifier.IdentifiedEMFObjectLocator;
import org.eclipse.emf.diffmerge.patch.runtime.modelaccess.EMFModelAccess;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.xtend.lib.annotations.AccessorType;
import org.eclipse.xtend.lib.annotations.Accessors;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Pure;

@SuppressWarnings("all")
public class EMFPatchApplication implements PatchApplication {
  @Accessors(AccessorType.PUBLIC_GETTER)
  private final EMFModelAccess modelAccess;
  
  @Accessors(AccessorType.PUBLIC_GETTER)
  private final ResourceSet input;
  
  @Accessors({ AccessorType.PUBLIC_GETTER, AccessorType.PROTECTED_SETTER })
  private IdentifiedEMFObjectLocator locator;
  
  private final ModelPatch modelPatch;
  
  private final PatchApplicationDiagnostic diagnostics;
  
  private final Map<String, EObject> storage;
  
  private final Multimap<FeatureHolder, EObject> addedOpposites = HashMultimap.<FeatureHolder, EObject>create();
  
  private final Multimap<FeatureHolder, EObject> removedOpposites = HashMultimap.<FeatureHolder, EObject>create();
  
  protected EMFPatchApplication(final ModelPatch modelPatch, final EMFModelAccess modelAccess, final ResourceSet input) {
    Preconditions.<ModelPatch>checkNotNull(modelPatch, "Model patch cannot be null");
    Preconditions.<EMFModelAccess>checkNotNull(modelAccess, "Model access cannot be null");
    Preconditions.<ResourceSet>checkNotNull(input, "Input cannot be null");
    this.modelPatch = modelPatch;
    this.modelAccess = modelAccess;
    this.input = input;
    PatchApplicationDiagnostic _patchApplicationDiagnostic = new PatchApplicationDiagnostic();
    this.diagnostics = _patchApplicationDiagnostic;
    this.storage = CollectionLiterals.<String, EObject>newHashMap();
    IdentifiedEMFObjectLocator _identifiedEMFObjectLocator = new IdentifiedEMFObjectLocator();
    this.locator = _identifiedEMFObjectLocator;
  }
  
  protected void apply() {
    final List<ModelPatchEntry> entries = this.modelPatch.getEntries();
    for (final ModelPatchEntry entry : entries) {
      try {
        final String contextId = entry.getContext().getIdentifier();
        EntryType _entryType = entry.getEntryType();
        if (_entryType != null) {
          switch (_entryType) {
            case ELEMENT:
              this.applyElementEntry(contextId, entry);
              break;
            case REFERENCE:
              this.applyReferenceEntry(contextId, entry);
              break;
            case ATTRIBUTE:
              this.applyAttributeEntry(contextId, entry);
              break;
            default:
              break;
          }
        }
      } catch (final Throwable _t) {
        if (_t instanceof Exception) {
          final Exception ex = (Exception)_t;
          final ModelPatchDiagnosticElement diagnosticElement = new ModelPatchDiagnosticElement();
          diagnosticElement.setCaughtException(ex);
          diagnosticElement.setProblematicEntry(entry);
          diagnosticElement.setMessage(ex.getMessage());
          this.diagnostics.addElement(diagnosticElement);
        } else {
          throw Exceptions.sneakyThrow(_t);
        }
      }
    }
  }
  
  private void applyElementEntry(final String elementId, final ModelPatchEntry entry) {
    final ElementEntry elementEntry = ((ElementEntry) entry);
    final String typeId = elementEntry.getType().getIdentifier();
    final Optional<EClassifier> locatedType = this.locator.locateEClassifier(typeId);
    ChangeDirection _direction = entry.getDirection();
    boolean _equals = Objects.equal(_direction, ChangeDirection.ADD);
    if (_equals) {
      boolean _isPresent = locatedType.isPresent();
      boolean _not = (!_isPresent);
      if (_not) {
        StringConcatenation _builder = new StringConcatenation();
        _builder.append("Could not find EClassifier");
        throw new IllegalStateException(_builder.toString());
      }
      final EClassifier type = locatedType.get();
      final Optional<EObject> existingId = this.findEObjectById(elementId);
      boolean _isPresent_1 = existingId.isPresent();
      if (_isPresent_1) {
        StringConcatenation _builder_1 = new StringConcatenation();
        _builder_1.append("Cannot create element with existing id");
        throw new IllegalStateException(_builder_1.toString());
      }
      final EObject createdElement = this.createElement(elementId, type);
      this.storage.put(elementId, createdElement);
    } else {
      final Optional<EObject> element = this.findEObjectById(elementId);
      boolean _isPresent_2 = element.isPresent();
      if (_isPresent_2) {
        this.modelAccess.remove(element.get());
      }
      final EObject eObject = this.storage.remove(elementId);
      boolean _equals_1 = Objects.equal(eObject, null);
      if (_equals_1) {
        StringConcatenation _builder_2 = new StringConcatenation();
        _builder_2.append("Could not find element in storage");
        throw new IllegalStateException(_builder_2.toString());
      }
    }
  }
  
  private void applyReferenceEntry(final String contextId, final ModelPatchEntry entry) {
    final Optional<EObject> locatedContext = this.findEObjectById(contextId);
    final ReferenceEntry referenceEntry = ((ReferenceEntry) entry);
    final String featureId = referenceEntry.getFeature().getIdentifier();
    final Optional<EStructuralFeature> locatedEStructuralFeature = this.locator.locateEStructuralFeature(featureId);
    boolean _isPresent = locatedContext.isPresent();
    boolean _not = (!_isPresent);
    if (_not) {
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("Could not find owner object");
      throw new IllegalStateException(_builder.toString());
    }
    final EObject owner = locatedContext.get();
    boolean _isPresent_1 = locatedEStructuralFeature.isPresent();
    boolean _not_1 = (!_isPresent_1);
    if (_not_1) {
      StringConcatenation _builder_1 = new StringConcatenation();
      _builder_1.append("Could not find feature");
      throw new IllegalStateException(_builder_1.toString());
    }
    final EStructuralFeature feature = locatedEStructuralFeature.get();
    if ((!(feature instanceof EReference))) {
      StringConcatenation _builder_2 = new StringConcatenation();
      _builder_2.append("Feature is not a reference");
      throw new IllegalStateException(_builder_2.toString());
    }
    final EReference reference = ((EReference) feature);
    final String targetId = referenceEntry.getTarget().getIdentifier();
    EObject target = this.storage.get(targetId);
    boolean _equals = Objects.equal(target, null);
    if (_equals) {
      final Optional<EObject> locatedTargetObject = this.findEObjectById(targetId);
      target = locatedTargetObject.orNull();
    }
    boolean _equals_1 = Objects.equal(target, null);
    if (_equals_1) {
      StringConcatenation _builder_3 = new StringConcatenation();
      _builder_3.append("Could not find target object");
      throw new IllegalStateException(_builder_3.toString());
    }
    final Optional<Integer> index = referenceEntry.getIndex();
    ChangeDirection _direction = referenceEntry.getDirection();
    boolean _equals_2 = Objects.equal(_direction, ChangeDirection.ADD);
    if (_equals_2) {
      this.setOrAdd(owner, reference, target, index);
    } else {
      this.unsetOrRemove(owner, reference, target, index);
      boolean _isContainment = reference.isContainment();
      if (_isContainment) {
        this.storage.put(targetId, target);
      }
    }
  }
  
  private void applyAttributeEntry(final String elementId, final ModelPatchEntry entry) {
    final Optional<EObject> locatedEObject = this.findEObjectById(elementId);
    final AttributeEntry attributeEntry = ((AttributeEntry) entry);
    final String featureId = attributeEntry.getFeature().getIdentifier();
    final Optional<EStructuralFeature> locatedEStructuralFeature = this.locator.locateEStructuralFeature(featureId);
    boolean _isPresent = locatedEObject.isPresent();
    boolean _not = (!_isPresent);
    if (_not) {
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("Could not find owner object");
      throw new IllegalStateException(_builder.toString());
    }
    boolean _isPresent_1 = locatedEStructuralFeature.isPresent();
    boolean _not_1 = (!_isPresent_1);
    if (_not_1) {
      StringConcatenation _builder_1 = new StringConcatenation();
      _builder_1.append("Could not find feature");
      throw new IllegalStateException(_builder_1.toString());
    }
    final EObject owner = locatedEObject.get();
    final EStructuralFeature feature = locatedEStructuralFeature.get();
    Object value = null;
    if ((feature instanceof EAttribute)) {
      final EDataType attributeType = ((EAttribute) feature).getEAttributeType();
      value = EcoreUtil.createFromString(attributeType, attributeEntry.getValue());
    } else {
      StringConcatenation _builder_2 = new StringConcatenation();
      _builder_2.append("Feature is not an attribute");
      throw new IllegalStateException(_builder_2.toString());
    }
    final Optional<Integer> index = attributeEntry.getIndex();
    ChangeDirection _direction = entry.getDirection();
    boolean _equals = Objects.equal(_direction, ChangeDirection.ADD);
    if (_equals) {
      this.setOrAdd(owner, feature, value, index);
    } else {
      this.unsetOrRemove(owner, feature, value, index);
    }
  }
  
  private EObject createElement(final String elementId, final EClassifier type) {
    final EObject eObject = this.modelAccess.create(type);
    EcoreUtil.setID(eObject, elementId);
    return eObject;
  }
  
  private void setOrAdd(final EObject owner, final EStructuralFeature feature, final Object value, final Optional<Integer> index) {
    final FeatureHolder holder = new FeatureHolder(owner, feature);
    if ((this.addedOpposites.containsKey(holder) && this.addedOpposites.get(holder).contains(value))) {
      this.modelAccess.changeIndex(owner, feature, value, index);
    } else {
      this.modelAccess.add(owner, feature, value, index);
      if ((feature instanceof EReference)) {
        final EReference opposite = ((EReference)feature).getEOpposite();
        boolean _notEquals = (!Objects.equal(opposite, null));
        if (_notEquals) {
          final FeatureHolder oppositeHolder = new FeatureHolder(((EObject) value), opposite);
          this.addedOpposites.put(oppositeHolder, owner);
        }
      }
    }
  }
  
  private void unsetOrRemove(final EObject owner, final EStructuralFeature feature, final Object value, final Optional<Integer> index) {
    final FeatureHolder holder = new FeatureHolder(owner, feature);
    if ((this.removedOpposites.containsKey(holder) && this.removedOpposites.get(holder).contains(value))) {
      return;
    } else {
      this.modelAccess.remove(owner, feature, value, index);
      if ((feature instanceof EReference)) {
        final EReference opposite = ((EReference)feature).getEOpposite();
        boolean _notEquals = (!Objects.equal(opposite, null));
        if (_notEquals) {
          final FeatureHolder oppositeHolder = new FeatureHolder(((EObject) value), opposite);
          this.removedOpposites.put(oppositeHolder, owner);
        }
      }
    }
  }
  
  private Optional<EObject> findEObjectById(final String identifier) {
    Preconditions.<String>checkNotNull(identifier, "Identifier cannot be null");
    EObject eObject = this.storage.get(identifier);
    boolean _equals = Objects.equal(eObject, null);
    if (_equals) {
      final Optional<EObject> locatedEObject = this.locator.locateEObject(identifier);
      boolean _isPresent = locatedEObject.isPresent();
      if (_isPresent) {
        return locatedEObject;
      }
    }
    return Optional.<EObject>fromNullable(eObject);
  }
  
  @Override
  public PatchApplicationDiagnostic getDiagnostics() {
    return this.diagnostics;
  }
  
  @Override
  public ModelPatch getModelPatch() {
    return this.modelPatch;
  }
  
  @Pure
  public EMFModelAccess getModelAccess() {
    return this.modelAccess;
  }
  
  @Pure
  public ResourceSet getInput() {
    return this.input;
  }
  
  @Pure
  public IdentifiedEMFObjectLocator getLocator() {
    return this.locator;
  }
  
  protected void setLocator(final IdentifiedEMFObjectLocator locator) {
    this.locator = locator;
  }
}
